@@ -45,19 +45,20 @@
 #include "idle-handles.h"
 #include "idle-im-manager.h"
 #include "idle-muc-manager.h"
+#include "idle-roomlist-manager.h"
 #include "idle-parser.h"
 #include "idle-server-connection.h"
 
 #include "extensions/extensions.h"    /* Renaming */
 
+#define DEFAULT_KEEPALIVE_INTERVAL 30 /* sec */
+#define MISSED_KEEPALIVES_BEFORE_DISCONNECTING 3
+
 /* From RFC 2813 :
  * This in essence means that the client may send one (1) message every
  * two (2) seconds without being adversely affected.  Services MAY also
  * be subject to this mechanism.
  */
-
-#define DEFAULT_KEEPALIVE_INTERVAL 30 /* sec */
-
 #define MSG_QUEUE_TIMEOUT 2
 static gboolean flush_queue_faster = FALSE;
 
@@ -139,16 +140,18 @@ enum {
 	LAST_PROPERTY_ENUM
 };
 
-/* private structure */
-typedef struct _IdleConnectionPrivate IdleConnectionPrivate;
 struct _IdleConnectionPrivate {
 	/*
 	 * network connection
 	 */
-
 	IdleServerConnection *conn;
 	guint sconn_status;
 
+	/* When we sent a PING to the server which it hasn't PONGed for yet, or 0 if
+	 * there isn't a PING outstanding.
+	 */
+	gint64 ping_time;
+
 	/* IRC connection properties */
 	char *nickname;
 	char *server;
@@ -198,8 +201,6 @@ struct _IdleConnectionPrivate {
 	TpSimplePasswordManager *password_manager;
 };
 
-#define IDLE_CONNECTION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), IDLE_TYPE_CONNECTION, IdleConnectionPrivate))
-
 static void _iface_create_handle_repos(TpBaseConnection *self, TpHandleRepoIface **repos);
 static GPtrArray *_iface_create_channel_managers(TpBaseConnection *self);
 static gchar *_iface_get_unique_connection_name(TpBaseConnection *self);
@@ -212,6 +213,8 @@ static IdleParserHandlerResult _erroneous_nickname_handler(IdleParser *parser, I
 static IdleParserHandlerResult _nick_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data);
 static IdleParserHandlerResult _nickname_in_use_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data);
 static IdleParserHandlerResult _ping_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data);
+static IdleParserHandlerResult _pong_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data);
+static IdleParserHandlerResult _unknown_command_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data);
 static IdleParserHandlerResult _version_privmsg_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data);
 static IdleParserHandlerResult _welcome_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data);
 static IdleParserHandlerResult _whois_user_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data);
@@ -224,7 +227,7 @@ static void send_quit_request(IdleConnection *conn);
 static void connection_connect_cb(IdleConnection *conn, gboolean success, TpConnectionStatusReason fail_reason);
 static void connection_disconnect_cb(IdleConnection *conn, TpConnectionStatusReason reason);
 static gboolean idle_connection_hton(IdleConnection *obj, const gchar *input, gchar **output, GError **_error);
-static void idle_connection_ntoh(IdleConnection *obj, const gchar *input, gchar **output);
+static gchar *idle_connection_ntoh(IdleConnection *obj, const gchar *input);
 
 static void idle_connection_add_queue_timeout (IdleConnection *self);
 static void idle_connection_clear_queue_timeout (IdleConnection *self);
@@ -236,8 +239,9 @@ static void conn_aliasing_fill_contact_attributes (
     GHashTable *attributes_hash);
 
 static void idle_connection_init(IdleConnection *obj) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(obj);
+	IdleConnectionPrivate *priv = G_TYPE_INSTANCE_GET_PRIVATE (obj, IDLE_TYPE_CONNECTION, IdleConnectionPrivate);
 
+	obj->priv = priv;
 	priv->sconn_status = SERVER_CONNECTION_STATE_NOT_CONNECTED;
 	priv->msg_queue = g_queue_new();
 
@@ -258,7 +262,8 @@ idle_connection_constructed (GObject *object)
 }
 
 static void idle_connection_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(obj);
+	IdleConnection *self = IDLE_CONNECTION (obj);
+	IdleConnectionPrivate *priv = self->priv;
 
 	switch (prop_id) {
 		case PROP_NICKNAME:
@@ -319,7 +324,8 @@ static void idle_connection_set_property(GObject *obj, guint prop_id, const GVal
 }
 
 static void idle_connection_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(obj);
+	IdleConnection *self = IDLE_CONNECTION (obj);
+	IdleConnectionPrivate *priv = self->priv;
 
 	switch (prop_id) {
 		case PROP_NICKNAME:
@@ -374,7 +380,7 @@ static void idle_connection_get_property(GObject *obj, guint prop_id, GValue *va
 
 static void idle_connection_dispose (GObject *object) {
 	IdleConnection *self = IDLE_CONNECTION(object);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(self);
+	IdleConnectionPrivate *priv = self->priv;
 
 	if (priv->dispose_has_run)
 		return;
@@ -407,7 +413,8 @@ static void idle_connection_dispose (GObject *object) {
 }
 
 static void idle_connection_finalize (GObject *object) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(object);
+	IdleConnection *self = IDLE_CONNECTION (object);
+	IdleConnectionPrivate *priv = self->priv;
 	IdleOutputPendingMsg *msg;
 
 	idle_contact_info_finalize(object);
@@ -508,8 +515,9 @@ static void idle_connection_class_init(IdleConnectionClass *klass) {
 		flush_queue_faster = TRUE;
 }
 
-static GPtrArray *_iface_create_channel_managers(TpBaseConnection *self) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(self);
+static GPtrArray *_iface_create_channel_managers(TpBaseConnection *base) {
+	IdleConnection *self = IDLE_CONNECTION (base);
+	IdleConnectionPrivate *priv = self->priv;
 	GPtrArray *managers = g_ptr_array_sized_new(1);
 	GObject *manager;
 
@@ -519,9 +527,12 @@ static GPtrArray *_iface_create_channel_managers(TpBaseConnection *self) {
 	manager = g_object_new(IDLE_TYPE_MUC_MANAGER, "connection", self, NULL);
 	g_ptr_array_add(managers, manager);
 
-	priv->password_manager = tp_simple_password_manager_new(self);
+	priv->password_manager = tp_simple_password_manager_new(base);
 	g_ptr_array_add(managers, priv->password_manager);
 
+	manager = g_object_new(IDLE_TYPE_ROOMLIST_MANAGER, "connection", self, NULL);
+	g_ptr_array_add(managers, manager);
+
 	return managers;
 }
 
@@ -532,15 +543,17 @@ static void _iface_create_handle_repos(TpBaseConnection *self, TpHandleRepoIface
 	idle_handle_repos_init(repos);
 }
 
-static gchar *_iface_get_unique_connection_name(TpBaseConnection *self) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(self);
+static gchar *_iface_get_unique_connection_name(TpBaseConnection *base) {
+	IdleConnection *self = IDLE_CONNECTION (base);
+	IdleConnectionPrivate *priv = self->priv;
 
 	return g_strdup_printf("%s@%s%p", priv->nickname, priv->server, self);
 }
 
 static gboolean _finish_shutdown_idle_func(gpointer data) {
 	TpBaseConnection *conn = TP_BASE_CONNECTION(data);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnection *self = IDLE_CONNECTION(conn);
+	IdleConnectionPrivate *priv = self->priv;
 	if (priv->force_disconnect_id != 0) {
 		g_source_remove(priv->force_disconnect_id);
 	}
@@ -554,14 +567,16 @@ static gboolean
 _force_disconnect (gpointer data)
 {
 	IdleConnection *conn = IDLE_CONNECTION(data);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
-	idle_server_connection_disconnect_async(priv->conn, NULL, NULL, NULL);
+	IdleConnectionPrivate *priv = conn->priv;
+
+	IDLE_DEBUG("gave up waiting, forcibly disconnecting");
+	idle_server_connection_force_disconnect(priv->conn);
 	return FALSE;
 }
 
 static void _iface_disconnected(TpBaseConnection *self) {
 	IdleConnection *conn = IDLE_CONNECTION(self);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 
 	/* we never got around to actually creating the connection
 	 * iface object because we were still trying to connect, so
@@ -580,8 +595,9 @@ static void _iface_disconnected(TpBaseConnection *self) {
 	priv->force_disconnect_id = g_timeout_add_seconds(2, _force_disconnect, conn);
 }
 
-static void _iface_shut_down(TpBaseConnection *self) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(self);
+static void _iface_shut_down(TpBaseConnection *base) {
+	IdleConnection *self = IDLE_CONNECTION (base);
+	IdleConnectionPrivate *priv = self->priv;
 
 	if (priv->quitting)
 		return;
@@ -590,7 +606,7 @@ static void _iface_shut_down(TpBaseConnection *self) {
 	 * iface object because we were still trying to connect, so
 	 * don't try to send any traffic down it */
 	if (priv->conn == NULL) {
-		g_idle_add(_finish_shutdown_idle_func, self);;
+		g_idle_add(_finish_shutdown_idle_func, self);
 	} else {
 		idle_server_connection_disconnect_async(priv->conn, NULL, NULL, NULL);
 	}
@@ -624,7 +640,7 @@ static void _start_connecting_continue(IdleConnection *conn);
 static void _password_prompt_cb(GObject *source, GAsyncResult *result, gpointer user_data) {
 	IdleConnection *conn = user_data;
 	TpBaseConnection *base_conn = TP_BASE_CONNECTION(conn);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	const GString *password;
 	GError *error = NULL;
 
@@ -651,7 +667,7 @@ static void _password_prompt_cb(GObject *source, GAsyncResult *result, gpointer
 
 static gboolean _iface_start_connecting(TpBaseConnection *self, GError **error) {
 	IdleConnection *conn = IDLE_CONNECTION(self);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 
 	g_assert(priv->nickname != NULL);
 	g_assert(priv->server != NULL);
@@ -675,7 +691,7 @@ static gboolean _iface_start_connecting(TpBaseConnection *self, GError **error)
 static void _connection_connect_ready(GObject *source_object, GAsyncResult *res, gpointer user_data) {
 	IdleServerConnection *sconn = IDLE_SERVER_CONNECTION(source_object);
 	IdleConnection *conn = IDLE_CONNECTION(user_data);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	GError *error = NULL;
 
 	if (!idle_server_connection_connect_finish(sconn, res, &error)) {
@@ -697,6 +713,8 @@ static void _connection_connect_ready(GObject *source_object, GAsyncResult *res,
 	idle_parser_add_handler(conn->parser, IDLE_PARSER_NUMERIC_WHOISUSER, _whois_user_handler, conn);
 
 	idle_parser_add_handler(conn->parser, IDLE_PARSER_CMD_PING, _ping_handler, conn);
+	idle_parser_add_handler(conn->parser, IDLE_PARSER_PREFIXCMD_PONG, _pong_handler, conn);
+	idle_parser_add_handler(conn->parser, IDLE_PARSER_NUMERIC_UNKNOWNCOMMAND, _unknown_command_handler, conn);
 
 	idle_parser_add_handler_with_priority(conn->parser, IDLE_PARSER_PREFIXCMD_NICK, _nick_handler, conn, IDLE_PARSER_HANDLER_PRIORITY_FIRST);
 	idle_parser_add_handler(conn->parser, IDLE_PARSER_PREFIXCMD_PRIVMSG_USER, _version_privmsg_handler, conn);
@@ -705,7 +723,7 @@ static void _connection_connect_ready(GObject *source_object, GAsyncResult *res,
 }
 
 static void _start_connecting_continue(IdleConnection *conn) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	IdleServerConnection *sconn;
 
 	if (tp_str_empty(priv->realname)) {
@@ -736,7 +754,7 @@ static void _start_connecting_continue(IdleConnection *conn) {
 static gboolean keepalive_timeout_cb(gpointer user_data);
 
 static void sconn_status_changed_cb(IdleServerConnection *sconn, IdleServerConnectionState state, IdleServerConnectionStateReason reason, IdleConnection *conn) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	TpConnectionStatusReason tp_reason;
 
 	/* cancel scheduled forced disconnect since we are now disconnected */
@@ -792,9 +810,7 @@ static void sconn_status_changed_cb(IdleServerConnection *sconn, IdleServerConne
 }
 
 static void sconn_received_cb(IdleServerConnection *sconn, gchar *raw_msg, IdleConnection *conn) {
-	gchar *converted;
-
-	idle_connection_ntoh(conn, raw_msg, &converted);
+	gchar *converted = idle_connection_ntoh(conn, raw_msg);
 	idle_parser_receive(conn->parser, converted);
 
 	g_free(converted);
@@ -802,17 +818,41 @@ static void sconn_received_cb(IdleServerConnection *sconn, gchar *raw_msg, IdleC
 
 static gboolean keepalive_timeout_cb(gpointer user_data) {
 	IdleConnection *conn = IDLE_CONNECTION(user_data);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	gchar cmd[IRC_MSG_MAXLEN + 1];
-	gint64 ping_time;
+	gint64 now;
 
-	if (priv->sconn_status != SERVER_CONNECTION_STATE_CONNECTED) {
+	if (priv->sconn_status != SERVER_CONNECTION_STATE_CONNECTED ||
+	    priv->quitting) {
 		priv->keepalive_timeout = 0;
 		return FALSE;
 	}
 
-	ping_time = g_get_real_time();
-	g_snprintf(cmd, IRC_MSG_MAXLEN + 1, "PING %" G_GINT64_FORMAT, ping_time);
+	now = g_get_real_time();
+
+	if (priv->ping_time != 0) {
+		gint64 seconds_since_ping = (now - priv->ping_time) / G_USEC_PER_SEC;
+		gint64 grace_period = priv->keepalive_interval * MISSED_KEEPALIVES_BEFORE_DISCONNECTING;
+
+		if (seconds_since_ping > grace_period) {
+			IDLE_DEBUG("haven't heard from the server in %" G_GINT64_FORMAT " seconds "
+				"(more than %u keepalive intervals)",
+				seconds_since_ping, MISSED_KEEPALIVES_BEFORE_DISCONNECTING);
+
+			idle_server_connection_force_disconnect(priv->conn);
+			return FALSE;
+		}
+
+		return TRUE;
+	}
+
+	if (priv->msg_queue->length > 0) {
+		/* No point in sending a PING if we're sending data anyway. */
+		return TRUE;
+	}
+
+	priv->ping_time = now;
+	g_snprintf(cmd, IRC_MSG_MAXLEN + 1, "PING %" G_GINT64_FORMAT, priv->ping_time);
 	_send_with_priority(conn, cmd, SERVER_CMD_MIN_PRIORITY);
 
 	return TRUE;
@@ -821,7 +861,7 @@ static gboolean keepalive_timeout_cb(gpointer user_data) {
 static void _msg_queue_timeout_ready(GObject *source_object, GAsyncResult *res, gpointer user_data) {
 	IdleServerConnection *sconn = IDLE_SERVER_CONNECTION(source_object);
 	IdleConnection *conn = IDLE_CONNECTION (user_data);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	GError *error = NULL;
 
 	priv->msg_sending = FALSE;
@@ -837,7 +877,7 @@ static void _msg_queue_timeout_ready(GObject *source_object, GAsyncResult *res,
 
 static gboolean msg_queue_timeout_cb(gpointer user_data) {
 	IdleConnection *conn = IDLE_CONNECTION(user_data);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	IdleOutputPendingMsg *output_msg;
 
 	IDLE_DEBUG("called");
@@ -860,9 +900,9 @@ static gboolean msg_queue_timeout_cb(gpointer user_data) {
 		return FALSE;
 	}
 
+	priv->msg_sending = TRUE;
 	idle_server_connection_send_async(priv->conn, output_msg->message, NULL, _msg_queue_timeout_ready, conn);
 	idle_output_pending_msg_free (output_msg);
-	priv->msg_sending = TRUE;
 
 	return TRUE;
 }
@@ -870,7 +910,7 @@ static gboolean msg_queue_timeout_cb(gpointer user_data) {
 static void
 idle_connection_add_queue_timeout (IdleConnection *self)
 {
-  IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE (self);
+  IdleConnectionPrivate *priv = self->priv;
 
   if (priv->msg_queue_timeout == 0)
     {
@@ -894,7 +934,7 @@ idle_connection_add_queue_timeout (IdleConnection *self)
 static void
 idle_connection_clear_queue_timeout (IdleConnection *self)
 {
-  IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE (self);
+  IdleConnectionPrivate *priv = self->priv;
 
   if (priv->msg_queue_timeout != 0)
     {
@@ -907,7 +947,7 @@ idle_connection_clear_queue_timeout (IdleConnection *self)
  * Queue a IRC command for sending, clipping it to IRC_MSG_MAXLEN bytes and appending the required <CR><LF> to it
  */
 static void _send_with_priority(IdleConnection *conn, const gchar *msg, guint priority) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	gchar cmd[IRC_MSG_MAXLEN + 3];
 	int len;
 	gchar *converted;
@@ -947,7 +987,7 @@ void idle_connection_send(IdleConnection *conn, const gchar *msg) {
 gsize
 idle_connection_get_max_message_length(IdleConnection *conn)
 {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	if (priv->relay_prefix != NULL) {
 		/* server will add ':<relay_prefix> ' to all messages it relays on to
 		 * other users.  the +2 is for the initial : and the trailing space */
@@ -1056,6 +1096,35 @@ static IdleParserHandlerResult _ping_handler(IdleParser *parser, IdleParserMessa
 	return IDLE_PARSER_HANDLER_RESULT_HANDLED;
 }
 
+static IdleParserHandlerResult _pong_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data) {
+	IdleConnection *conn = IDLE_CONNECTION(user_data);
+	IdleConnectionPrivate *priv = conn->priv;
+
+	/* We could compare ping_time to g_get_real_time() and give some indication
+	 * of lag, if we were feeling enthusiastic.
+	 */
+	priv->ping_time = 0;
+
+	return IDLE_PARSER_HANDLER_RESULT_HANDLED;
+}
+
+static IdleParserHandlerResult _unknown_command_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data) {
+	IdleConnection *conn = IDLE_CONNECTION(user_data);
+	IdleConnectionPrivate *priv = conn->priv;
+	const gchar *command = g_value_get_string(g_value_array_get_nth(args, 0));
+
+	if (!tp_strdiff(command, "PING")) {
+		IDLE_DEBUG("PING not supported, disabling keepalive.");
+		g_source_remove(priv->keepalive_timeout);
+		priv->keepalive_timeout = 0;
+		priv->ping_time = 0;
+
+		return IDLE_PARSER_HANDLER_RESULT_HANDLED;
+	}
+
+	return IDLE_PARSER_HANDLER_RESULT_NOT_HANDLED;
+}
+
 static IdleParserHandlerResult _version_privmsg_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data) {
 	IdleConnection *conn = IDLE_CONNECTION(user_data);
 	const gchar *msg = g_value_get_string(g_value_array_get_nth(args, 2));
@@ -1092,7 +1161,7 @@ static IdleParserHandlerResult
 _whois_user_handler(IdleParser *parser, IdleParserMessageCode code, GValueArray *args, gpointer user_data)
 {
 	IdleConnection *conn = IDLE_CONNECTION(user_data);
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 
 	/* message format: <nick> <user> <host> * :<real name> */
 	TpHandle handle = g_value_get_uint(g_value_array_get_nth(args, 0));
@@ -1120,7 +1189,7 @@ static void irc_handshakes(IdleConnection *conn) {
 	g_assert(conn != NULL);
 	g_assert(IDLE_IS_CONNECTION(conn));
 
-	priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	priv = conn->priv;
 
 	if ((priv->password != NULL) && (priv->password[0] != '\0')) {
 		g_snprintf(msg, IRC_MSG_MAXLEN + 1, "PASS %s", priv->password);
@@ -1139,7 +1208,7 @@ static void irc_handshakes(IdleConnection *conn) {
 }
 
 static void send_quit_request(IdleConnection *conn) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 	gchar cmd[IRC_MSG_MAXLEN + 1] = "QUIT";
 
 	if (priv->quit_message != NULL)
@@ -1170,7 +1239,7 @@ static void connection_disconnect_cb(IdleConnection *conn, TpConnectionStatusRea
 
 static void
 _queue_alias_changed(IdleConnection *conn, TpHandle handle, const gchar *alias) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 
 	if (!priv->queued_aliases_owners) {
 		TpHandleRepoIface *handles = tp_base_connection_get_handles(TP_BASE_CONNECTION(conn), TP_HANDLE_TYPE_CONTACT);
@@ -1215,7 +1284,7 @@ void idle_connection_canon_nick_receive(IdleConnection *conn, TpHandle handle, c
 }
 
 void idle_connection_emit_queued_aliases_changed(IdleConnection *conn) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(conn);
+	IdleConnectionPrivate *priv = conn->priv;
 
 	if (!priv->queued_aliases)
 		return;
@@ -1369,7 +1438,7 @@ static void idle_connection_set_aliases(TpSvcConnectionInterfaceAliasing *iface,
 }
 
 static gboolean idle_connection_hton(IdleConnection *obj, const gchar *input, gchar **output, GError **_error) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(obj);
+	IdleConnectionPrivate *priv = obj->priv;
 	GError *error = NULL;
 	gsize bytes_written;
 	gchar *ret;
@@ -1393,16 +1462,47 @@ static gboolean idle_connection_hton(IdleConnection *obj, const gchar *input, gc
 	return TRUE;
 }
 
-static void idle_connection_ntoh(IdleConnection *obj, const gchar *input, gchar **output) {
-	IdleConnectionPrivate *priv = IDLE_CONNECTION_GET_PRIVATE(obj);
+#define U_FFFD_REPLACEMENT_CHARACTER_UTF8 "\357\277\275"
+
+static gchar *
+idle_salvage_utf8 (gchar *supposed_utf8, gssize bytes)
+{
+	GString *salvaged = g_string_sized_new (bytes);
+	const gchar *end;
+	gchar *ret;
+	gsize ret_len;
+
+	while (!g_utf8_validate (supposed_utf8, bytes, &end)) {
+		gssize valid_bytes = end - supposed_utf8;
+
+		g_string_append_len (salvaged, supposed_utf8, valid_bytes);
+		g_string_append_len (salvaged, U_FFFD_REPLACEMENT_CHARACTER_UTF8, 3);
+
+		supposed_utf8 += (valid_bytes + 1);
+		bytes -= (valid_bytes + 1);
+	}
+
+	g_string_append_len (salvaged, supposed_utf8, bytes);
+
+	ret_len = salvaged->len;
+	ret = g_string_free (salvaged, FALSE);
+
+	/* It had better be valid now… */
+	g_return_val_if_fail (g_utf8_validate (ret, ret_len, NULL), ret);
+	return ret;
+}
+
+
+static gchar *
+idle_connection_ntoh(IdleConnection *obj, const gchar *input) {
+	IdleConnectionPrivate *priv = obj->priv;
 	GError *error = NULL;
 	gsize bytes_written;
 	gchar *ret;
 	gchar *p;
 
 	if (input == NULL) {
-		*output = NULL;
-		return;
+		return NULL;
 	}
 
 	ret = g_convert(input, -1, "UTF-8", priv->charset, NULL, &bytes_written, &error);
@@ -1417,10 +1517,19 @@ static void idle_connection_ntoh(IdleConnection *obj, const gchar *input, gchar
 			if (*p & (1 << 7))
 				*p = '?';
 		}
+	} else if (!g_utf8_validate (ret, bytes_written, NULL)) {
+		/* Annoyingly g_convert(UTF-8, UTF-8) doesn't filter out well-formed
+		 * non-characters, so we have to do some further processing.
+		 */
+		gchar *salvaged;
+
+		IDLE_DEBUG("Invalid UTF-8, salvaging what we can...");
+		salvaged = idle_salvage_utf8(ret, bytes_written);
+		g_free(ret);
+		ret = salvaged;
 	}
 
-	*output = ret;
-	return;
+	return ret;
 }
 
 static void _aliasing_iface_init(gpointer g_iface, gpointer iface_data) {
