@@ -362,6 +362,23 @@ void idle_server_connection_disconnect_async(IdleServerConnection *conn, GCancel
 						     user_data);
 }
 
+void idle_server_connection_force_disconnect(IdleServerConnection *conn) {
+	/* Passing a cancelled cancellable to g_io_stream_close_async() stops us
+	 * waiting for a TCP-level reply from the server which we already know
+	 * has gone away. Quoth the docs for g_io_stream_close():
+	 *
+	 *     Cancelling a close will still leave the stream closed, but some
+	 *     streams can use a faster close that doesn't block to e.g. check
+	 *     errors.
+	 */
+	GCancellable *kill_me_now = g_cancellable_new();
+
+	g_cancellable_cancel(kill_me_now);
+	idle_server_connection_disconnect_full_async(conn,
+		SERVER_CONNECTION_STATE_REASON_ERROR, kill_me_now, NULL, NULL);
+	g_object_unref(kill_me_now);
+}
+
 void idle_server_connection_disconnect_full_async(IdleServerConnection *conn, guint reason, GCancellable *cancellable, GAsyncReadyCallback callback, gpointer user_data) {
 	IdleServerConnectionPrivate *priv = IDLE_SERVER_CONNECTION_GET_PRIVATE(conn);
 	GSimpleAsyncResult *result;
